"""This module contains functions to generate different types of profiles."""

from .gratings import blazed_grating, sinusoidal_amplitude_grating, sinusoidal_phase_grating
from .hermite_gaussian import gaussian, hermite_gaussian
from .laguerre_gaussian import laguerre_gaussian
from .lens import lens
from .shapes import checkerboard, circle, rectangle, square
from .spatial_coherence import gaussian_schell_model, schell_model
from .special import airy, sinc
